using System;
using System.IdentityModel.Tokens.Jwt;
using System.IO;
using System.Security.Claims;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;

namespace FitConnect.Security;

public class RsaEncryption : IEncryption {
    private readonly X509Certificate2? _certificate;
    private readonly ILogger? _logger;
    private readonly RSA _rsa;
    private RSA? _privateKey;
    private RSA? _publicKey;


    public RsaEncryption(X509Certificate2? certificate, ILogger? logger) {
        _logger = logger;
        _rsa = RSA.Create(4096);

        if (certificate != null) {
            _certificate = certificate;
            ImportCertificate(certificate);
        }
    }


    /// <summary>
    ///     Just for Proof of Concept
    /// </summary>
    /// <returns></returns>
    public string GetTestToken() {
        var handler = new JwtSecurityTokenHandler();
        var token = new SecurityTokenDescriptor {
            Issuer = "FitConnect",
            Audience = "FitConnect",
            EncryptingCredentials =
                new X509EncryptingCredentials(_certificate ??
                                              new X509Certificate2(CreateSelfSignedCertificate())),
            Subject = new ClaimsIdentity(new Claim[] {
                new("Content", "Unencrypted content")
            })
        };
        return handler.CreateEncodedJwt(token);
    }


    public void ImportCertificate(X509Certificate2 cert) {
        if (!CheckCertificate(cert)) throw new Exception("Invalid certificate");

        _publicKey = cert.GetRSAPublicKey();

        if ((_publicKey?.KeySize ?? 0) == 0)
            throw new Exception("Invalid certificate, no public key");

        if (cert.HasPrivateKey) _privateKey = cert.GetRSAPrivateKey();

        if (_privateKey != null)
            _logger?.LogInformation("Certificate with private key imported");
        else
            _logger?.LogInformation("Certificate has no private key");
    }

    /// <summary>
    ///     Import a public key from a PEM file
    /// </summary>
    /// <param name="certificatePath"></param>
    /// <param name="password">Password for the certificate</param>
    /// <exception cref="ArgumentException"></exception>
    /// <exception cref="Exception"></exception>
    public void ImportCertificate(string certificatePath, string password) {
        _logger?.LogInformation("Importing certificate {CertPath}", certificatePath);
        var cert =
            new X509Certificate2(certificatePath, password, X509KeyStorageFlags.MachineKeySet);

        if (!CheckCertificate(cert)) throw new ArgumentException("Certificate is not valid");

        var parameters = cert.PublicKey.GetRSAPublicKey()?.ExportParameters(true);
        if (parameters == null) throw new Exception("Could not get public key from certificate");

        _rsa.ImportParameters(parameters.Value);
    }

    public byte[] DecryptData(byte[] data) {
        return (_privateKey ?? _rsa).Decrypt(data, RSAEncryptionPadding.OaepSHA256);
    }

    public string DecryptData(string data) {
        return Encoding.UTF8.GetString(DecryptData(Convert.FromBase64String(data)));
    }

    public byte[] ExportPublicKey() {
        return _rsa.ExportRSAPublicKey();
    }

    public byte[] ExportPrivateKey() {
        return _rsa.ExportRSAPrivateKey();
    }

    public byte[] EncryptData(byte[] data) {
        return (_publicKey ?? _rsa).Encrypt(data, RSAEncryptionPadding.OaepSHA256);
    }

    public byte[] EncryptData(byte[] data, byte[] publicKey) {
        _logger?.LogInformation(
            "Encrypting data with public key: {}",
            Convert.ToBase64String(_rsa.ExportRSAPublicKey()));

        _rsa.ImportRSAPublicKey(publicKey, out var read);
        return _rsa.Encrypt(data, RSAEncryptionPadding.OaepSHA256);
    }

    /// <summary>
    ///     Creating a self signed certificate
    /// </summary>
    /// <param name="exportPath">Location for storing the certificate files</param>
    /// <returns></returns>
    /// <exception cref="Exception"></exception>
    public static X509Certificate2 CreateSelfSignedCertificate(string? exportPath = null) {
        var rsa = RSA.Create(4096);

        var req = new CertificateRequest("c=DE, cn=fitconnect.de",
            rsa, HashAlgorithmName.SHA256, RSASignaturePadding.Pkcs1);

        var cert = req.CreateSelfSigned(DateTimeOffset.Now.AddSeconds(-5),
            DateTimeOffset.Now.AddYears(5));

        if (cert.GetRSAPublicKey() == null)
            throw new Exception("Certificate does not contain a public key");

        if (cert.GetRSAPrivateKey() == null)
            throw new Exception("Certificate does not contain a private key");

        // Export the certificate to a PEM file, just for
        // additional extern testing
        if (exportPath != null) ExportCertificateToFile(exportPath, cert);

        return cert;
    }

    private static void ExportCertificateToFile(string exportPath, X509Certificate cert) {
        // Create PFX (PKCS #12) with private key
        File.WriteAllBytes($"{exportPath}/certificate.pfx",
            cert.Export(X509ContentType.Pfx, ""));

        // Create Base 64 encoded CER (public key only)
        File.WriteAllText($"{exportPath}/certificate.cer",
            "-----BEGIN CERTIFICATE-----\r\n"
            + Convert.ToBase64String(cert.Export(X509ContentType.Cert, ""),
                Base64FormattingOptions.InsertLineBreaks)
            + "\r\n-----END CERTIFICATE-----");
    }


    /// <summary>
    ///     Checking the certificate for validity
    /// </summary>
    /// <param name="cert">Certificate to check</param>
    /// <returns>true for a fully approved certificate</returns>
    private bool CheckCertificate(X509Certificate2 cert) {
        return true;
    }
}
